<?php
/**
 * Plugin Name: Flywire for WooCommerce
 * Plugin URI: http://www.flywire.com/
 * Description: Enable Flywire payments option for WooCommerce
 * Author: Flywire
 * Author URI: http://www.flywire.com/
 * Version: 1.0.3
 * Text Domain: flywire-payment-gateway
 * Domain Path: /i18n/languages/
 *
 * Copyright: (c) 2021 Flywire Payments Corp and WooCommerce
 *
 * License: GNU General Public License v3.0
 * License URI: http://www.gnu.org/licenses/gpl-3.0.html
 *
 * @package   flywire-payment-gateway
 * @author    Flywire
 * @category  Admin
 * @copyright Copyright (c) 2021 Flywire Payments Corp and WooCommerce
 * @license   http://www.gnu.org/licenses/gpl-3.0.html GNU General Public License v3.0
 *
 * Enable Flywire payments option for WooCommerce
 */

defined( 'ABSPATH' ) or exit;


// Make sure WooCommerce is active
if ( ! in_array( 'woocommerce/woocommerce.php', apply_filters( 'active_plugins', get_option( 'active_plugins' ) ) ) ) {
	return;
}


/**
 * Add the gateway to WC Available Gateways
 *
 * @param array $gateways all available WC gateways
 *
 * @return array $gateways all WC gateways + offline gateway
 * @since 1.0.0
 */
function flywire_pay_add_to_gateways( $gateways ) {
	$gateways[] = 'WC_Gateway_Flywire';

	return $gateways;
}

add_filter( 'woocommerce_payment_gateways', 'flywire_pay_add_to_gateways' );


/**
 * Adds plugin page links
 *
 * @param array $links all plugin links
 *
 * @return array $links all plugin links + our custom links (i.e., "Settings")
 * @since 1.0.0
 */
function flywire_pay_gateway_plugin_links( $links ) {

	$plugin_links = array(
		'<a href="' . admin_url( 'admin.php?page=wc-settings&tab=checkout&section=flywire' ) . '">' . __( 'Configure', 'flywire-payment-gateway' ) . '</a>'
	);

	return array_merge( $plugin_links, $links );
}

add_filter( 'plugin_action_links_' . plugin_basename( __FILE__ ), 'flywire_pay_gateway_plugin_links' );

/**
 * Flywire for WooCommerce
 *
 * Provides a Flywire Payment Gateway
 *
 * @class        WC_Gateway_Flywire
 * @extends        WC_Payment_Gateway
 * @version        1.0.0
 * @package        WooCommerce/Classes/Payment
 * @author        Flywire
 */
add_action( 'plugins_loaded', 'flywire_pay_gateway_init', 11 );

function flywire_pay_gateway_init() {

	class WC_Gateway_Flywire extends WC_Payment_Gateway {

		/**
		 * Whether or not logging is enabled
		 *
		 * @var bool
		 */
		public static $log_enabled = false;

		/**
		 * Logger instance
		 *
		 * @var WC_Logger
		 */
		public static $log = false;

		/**
		 * Is in test mode.  If true, use UAT credentials and URLs.
		 *
		 * @var bool
		 */
		public $testmode;

		/**
		 * Flywire 3-letter portal code
		 *
		 * @var string
		 */
		public $portal;

		/**
		 * Label for payment button on Checkout
		 *
		 * @var string
		 */
		public $button_text;

		/**
		 * Instructions to add to email
		 *
		 * @var string
		 */
		public $instructions;

		/**
		 * Complete order for virtual products on payment success, if true
		 *
		 * @var bool
		 */
		public $complete_order;

		/**
		 * Field mapping definition object
		 *
		 * @var string
		 */
		public $field_mapping;

		/**
		 * Webhook URL
		 *
		 * @var string
		 */
		public $api_url;

		/**
         * Consider Flywire payment complete on Guaranteed instead of Delivered
         *
		 * @var bool
		 */
		public $on_guaranteed;

		/**
         * Add Flywire status updates as order notes
         *
		 * @var bool
		 */
		public $order_notes;

		/**
		 * Constructor for the gateway.
		 */
		public function __construct() {
			$this->id                 = 'flywire';
			$this->has_fields         = false;
			$this->method_title       = __( 'Flywire', 'flywire-payment-gateway' );
			$this->order_button_text  = $this->get_option( 'button_text' );
			$this->method_description = __( 'Redirect users to pay with Flywire.  Requires an existing Flywire account.  Please <a href="https://www.flywire.com/company/contact" target="_blank">contact us</a> for a free demo, more information, or to create an account.', 'flywire-payment-gateway' );
			$this->supports           = array(
				'products'
			);

			// Load the settings.
			$this->init_form_fields();
			$this->init_settings();

			// Define user set variables
			$this->title          = $this->get_option( 'title' );
			$this->description    = $this->get_option( 'description' );
			$this->enabled        = $this->get_option( 'enabled' );
			$this->testmode       = 'yes' === $this->get_option( 'testmode' );
			$this->complete_order = 'yes' === $this->get_option( 'complete_order' );
			$this->on_guaranteed  = 'yes' === $this->get_option( 'complete_on_guaranteed' );
			$this->portal         = $this->testmode ? $this->get_option( 'demo_portal' ) : $this->get_option( 'prod_portal' );
			$this->instructions   = $this->get_option( 'instructions' );
			$this->field_mapping  = $this->get_option( 'field_mapping' );
			$this->api_url        = WC()->api_request_url( 'WC_Gateway_Flywire' );
			$this->order_notes    = 'yes' === $this->get_option( 'enable_notes' );
			self::$log_enabled    = 'yes' === $this->get_option( 'enable_log' );

			// Actions
			add_action( 'woocommerce_update_options_payment_gateways_' . $this->id, array(
				$this,
				'process_admin_options'
			) );
			add_action( 'woocommerce_thankyou_' . $this->id, array( $this, 'thankyou_page' ) );
			add_action( 'woocommerce_email_before_order_table', array( $this, 'email_instructions' ), 10, 3 );
			add_action( 'wp_enqueue_scripts', array( $this, 'add_plugin_scripts' ) );
			add_action( 'woocommerce_after_checkout_form', array( $this, 'checkout_button' ) );

			if ( ! $this->is_valid_for_use() ) {
				WC_Gateway_Flywire::log( 'Flywire for WooCommerce is not valid for use' );
				$this->enabled = 'no';
			} else {
				include_once dirname( __FILE__ ) . '/includes/flywire-payment-gateway-callback-handler.php';
				new WC_Gateway_Flywire_Callback_Handler( $this );
			}
		}


		public function checkout_button() {
			?>
            <style>
                .invisible-button {
                    width: 1px;
                    height: 1px;
                    padding: 0;
                    border: none;
                    background-color: transparent;
                }
            </style>
            <button id="checkout-button" class="invisible-button" disabled></button>
			<?php
		}


		/**
		 * Initialize Gateway Settings Form Fields
		 */

		public function init_form_fields() {
			$this->form_fields = apply_filters( 'flywire_pay_form_fields', array(

				'enabled' => array(
					'title'   => __( 'Enable/Disable', 'flywire-payment-gateway' ),
					'type'    => 'checkbox',
					'label'   => __( 'Enable Flywire Payments', 'flywire-payment-gateway' ),
					'default' => 'yes'
				),

				'title' => array(
					'title'       => __( 'Title', 'flywire-payment-gateway' ),
					'type'        => 'text',
					'description' => __( 'This controls the title for the payment method the customer sees during checkout.', 'flywire-payment-gateway' ),
					'default'     => __( 'Flywire', 'flywire-payment-gateway' ),
					'desc_tip'    => true,
				),

				'description' => array(
					'title'       => __( 'Description', 'flywire-payment-gateway' ),
					'type'        => 'textarea',
					'description' => __( 'Payment method description that the customer will see on your checkout.', 'flywire-payment-gateway' ),
					'default'     => __( 'Secure payments by card, bank transfer or local payment options.  All with the lowest fees and around-the-clock multilingual support.', 'flywire-payment-gateway' ),
					'desc_tip'    => true,
				),

				'instructions' => array(
					'title'       => __( 'On-Hold Instructions', 'flywire-payment-gateway' ),
					'type'        => 'textarea',
					'description' => __( 'Instructions that will be added to the email for orders On Hold.', 'flywire-payment-gateway' ),
					'default'     => '',
					'desc_tip'    => true,
				),

				'button_text' => array(
					'title'       => __( 'Button Text', 'flywire-payment-gateway' ),
					'type'        => 'text',
					'description' => __( 'Label for the payment button on the Checkout page.', 'flywire-payment-gateway' ),
					'default'     => __( 'Pay with Flywire', 'flywire-payment-gateway' ),
					'desc_tip'    => true,
				),

				'field_mapping' => array(
					'title'       => __( 'Field Mapping', 'flywire-payment-gateway' ),
					'type'        => 'textarea',
					'description' => __( 'JSON configuration to define how WooCommerce fields will map to Flywire fields.  Should be supplied by your Flywire Relationship Manager or Solutions Consultant.', 'flywire-payment-gateway' ),
					'default'     => "{\r\n\t\"sender_country\":\"{{billing_country}}\",\r\n\t\"sender_first_name\":\"{{billing_first_name}}\",\r\n\t\"sender_last_name\":\"{{billing_last_name}}\",\r\n\t\"sender_address1\":\"{{billing_address_1}}\",\r\n\t\"sender_phone\":\"{{billing_phone}}\",\r\n\t\"sender_city\":\"{{billing_city}}\",\r\n\t\"sender_zip\":\"{{billing_postcode}}\",\r\n\t\"sender_email\":\"{{billing_email}}\",\r\n\t\"student_first_name\":\"{{billing_first_name}}\",\r\n\t\"student_last_name\":\"{{billing_last_name}}\",\r\n\t\"student_email\":\"{{billing_email}}\"\r\n}",
					'desc_tip'    => true,
				),

				'enable_notes' => array(
					'title'       => __( 'Verbose Order Notes', 'flywire-payment-gateway' ),
					'type'        => 'checkbox',
					'label'       => __( 'Add Flywire transaction status updates as notes on the WooCommerce order', 'flywire-payment-gateway' ),
					'default'     => 'yes',
				),

				'enable_log' => array(
					'title'       => __( 'Logging', 'flywire-payment-gateway' ),
					'type'        => 'checkbox',
					'label'       => __( 'Enable logging to file', 'flywire-payment-gateway' ),
					'description' => __( 'Logs development messages for debugging.  Turn off in Production to optimize.', 'flywire-payment-gateway' ),
					'default'     => 'no',
					'desc_tip'    => true,
				),

				'complete_on_guaranteed' => array(
					'title'       => __( 'Action on Guaranteed', 'flywire-payment-gateway' ),
					'type'        => 'checkbox',
					'label'       => __( 'Consider Flywire payment completed on Guaranteed status instead of Delivered', 'flywire-payment-gateway' ),
					'description' => __( 'If unchecked, will wait until Flywire "Delivered" status to consider payment complete.', 'flywire-payment-gateway' ),
					'default'     => 'yes',
					'desc_tip'    => true,
				),

				'complete_order' => array(
					'title'   => __( 'Complete Orders', 'flywire-payment-gateway' ),
					'type'    => 'checkbox',
					'label'   => __( 'Set WooCommerce order to complete on successful payment (virtual products only)', 'flywire-payment-gateway' ),
					'description' => __( 'If unchecked, WooCommerce orders must be manually set to completed after Flywire payment is complete.', 'flywire-payment-gateway' ),
					'default' => 'yes',
					'desc_tip'    => true,
				),

				'demo_details' => array(
					'title'       => __( 'Portal Info', 'flywire-payment-gateway' ),
					'type'        => 'title',
					'description' => __( 'Please contact your Relationship Manager if you need help determining these values.', 'flywire-payment-gateway' ),
				),

				'testmode' => array(
					'title'       => __( 'Test Mode', 'flywire-payment-gateway' ),
					'type'        => 'checkbox',
					'label'       => __( 'Use Demo environment', 'flywire-payment-gateway' ),
					'default'     => 'yes',
				),

				'demo_portal' => array(
					'title'       => __( 'Demo Portal Code', 'flywire-payment-gateway' ),
					'type'        => 'text',
					'description' => __( 'The Demo portal code for your institution.  A 3-letter code (e.g. FLC)', 'flywire-payment-gateway' ),
					'default'     => '',
					'desc_tip'    => true,
				),

				'prod_portal' => array(
					'title'       => __( 'Production Portal Code', 'flywire-payment-gateway' ),
					'type'        => 'text',
					'description' => __( 'The production portal code for your institution.  A 3-letter code (e.g. FLC)', 'flywire-payment-gateway' ),
					'default'     => '',
					'desc_tip'    => true,
				),
			) );
		}

		/**
		 * Add scripts for Flywire Embed 2.0
		 */
		public function add_plugin_scripts() {
			wp_enqueue_script( 'flywire_checkout', 'https://payment.flywire.com/assets/js/checkout.js', false, '1.0', false );
			wp_enqueue_script( 'flywire_checkout_config', plugins_url( 'assets/checkout_config.js', __FILE__ ), array( 'flywire_checkout' ), '1.0', true );
		}

		/**
		 * Processes and saves options.
		 * If there is an Exception thrown, will continue to save and validate fields, but will leave the erroring field out.
		 * If there is a validation error thrown, will display a message.
		 *
		 * @return bool was anything saved?
		 */
		public function process_admin_options(): bool {
			$has_validation_errors = false;
			$fields                = $this->get_form_fields();
			$this->init_settings();

			$post_data = $this->get_post_data();

			foreach ( $fields as $key => $field ) {
				if ( 'title' !== $this->get_field_type( $field ) ) {
					try {
						if ( method_exists( $this, 'validate_' . $key ) ) {
							try {
								$this->settings[ $key ] = call_user_func(
									array( $this, 'validate_' . $key ),
									$key,
									$this->get_field_value( $key, $field, $post_data ),
									$fields
								);
							} catch ( Exception $e ) {
								$has_validation_errors = true;
								$this->add_error( $e->getMessage() );
							}
						} else {
							$this->settings[ $key ] = $this->get_field_value( $key, $field, $post_data );
						}
					} catch ( Exception $e ) {
						$this->add_error( $e->getMessage() );
					}
				}
			}

			if ( $has_validation_errors ) {
				$this->display_errors();

				return false;
			} else {
				return update_option( $this->get_option_key(), apply_filters( 'woocommerce_settings_api_sanitized_fields_' . $this->id, $this->settings ), 'yes' );
			}
		}

		/**
		 * Validate the field_mapping setting, enforce valid JSON
		 *
		 * @param string $key The new setting key
		 * @param string $value The new setting value
		 * @param array $fields All form fields
		 *
		 * @return string
		 */
		protected function validate_field_mapping( string $key, string $value, array $fields ): string {
			if ( json_decode( $value ) === null ) {
				throw new Exception( '<strong>' . $fields[ $key ]['title'] . '</strong> is not valid JSON.  Please contact your Solutions Consultant to get the custom configuration for your institution.' );
			}

			return $value;
		}

		/**
		 * If test mode is enabled, enforce UAT portal required.  If not, enforce production portal required.
		 *
		 * @param string $key The new setting key
		 * @param string $value The new setting value
		 * @param array $fields All form fields
		 *
		 * @return string
		 */
		protected function validate_testmode( string $key, string $value, array $fields ): string {
			$post_data = $this->get_post_data();

			if ( 'yes' === $value && empty( $this->get_field_value( 'demo_portal', $fields['demo_portal'], $post_data ) ) ) {
				throw new Exception( 'If <strong>' . $fields[ $key ]['title'] . '</strong> is enabled, <strong>UAT ' . $fields['demo_portal']['title'] . '</strong> is required.' );
			} elseif ( 'no' === $value && empty( $this->get_field_value( 'prod_portal', $fields['prod_portal'], $post_data ) ) ) {
				throw new Exception( 'If <strong>' . $fields[ $key ]['title'] . '</strong> is disabled, <strong>Production ' . $fields['prod_portal']['title'] . '</strong> is required.' );
			}

			return $value;
		}

		/**
		 * Logging method.
		 *
		 * @param string $message Log message.
		 * @param string $level Optional. Default 'info'. Possible values:
		 *                      emergency|alert|critical|error|warning|notice|info|debug.
		 */
		public static function log( string $message, $level = 'info' ) {
			if ( self::$log_enabled ) {
				if ( empty( self::$log ) ) {
					self::$log = wc_get_logger();
				}
				self::$log->log( $level, $message, array( 'source' => 'flywire' ) );
			}
		}

		/**
		 * Get gateway icon.
		 *
		 * @return string
		 */
		public function get_icon(): string {
			$html = '';
			$html .= '<img src="' . plugins_url( 'assets/flywire_f.png', __FILE__ ) . '" alt="' . __( 'Flywire icon', 'flywire-payment-gateway' ) . '" />';
			$html .= sprintf( '<a href="%1$s" target="_blank" class="about_flywire">' . __( 'What is Flywire?', 'flywire-payment-gateway' ) . '</a>', esc_url( 'https://help.flywire.com/hc/en-us/articles/360012463014-What-is-Flywire-' ) );

			return apply_filters( 'woocommerce_gateway_icon', $html, $this->id );
		}


		/**
		 * Check if this gateway is valid for use
		 *
		 * @return bool
		 */
		public function is_valid_for_use(): bool {
			// Placeholder for future functionality to validate plugin setup, parameters, etc
			return true;
		}

		/**
		 * Add content to the on-hold emails.
		 *
		 * @access public
		 *
		 * @param WC_Order $order
		 * @param bool $sent_to_admin
		 * @param bool $plain_text
		 */
		public function email_instructions( WC_Order $order, bool $sent_to_admin, bool $plain_text = false ) {
			if ( $this->instructions && ! $sent_to_admin && $this->id === $order->get_payment_method() && $order->has_status( 'on-hold' ) ) {
				echo wpautop( wptexturize( $this->instructions ) ) . PHP_EOL;
			}
		}


		/**
		 * Custom content for the order received thank you page.
		 */
		public function thankyou_page() {
			echo wp_kses_post( wpautop( wptexturize( sprintf( __( '<p id="thank-you-msg">Thank you for paying with Flywire.  Should you need to contact <a href="http://www.flywire.com/support" target="_blank">Flywire Support</a> for any reason, please reference your Flywire payment ID: %s</p><p>Regards,<br/>The Flywire Team</p>', 'flywire-payment-gateway' ), $_GET['reference'] ) ) ) );
		}

		/**
		 * Process the payment and return the result
		 *
		 * @param int $order_id WC Order ID
		 * @return array
		 */
		public function process_payment( $order_id ): array {
			include_once dirname( __FILE__ ) . '/includes/flywire-payment-gateway-request.php';
			$FW_request = new WC_Gateway_Flywire_Request( $this, $order_id );

			try {
			    $config= $FW_request->get_config();
			} catch ( Exception $e ) {
				wc_add_notice( $e->getMessage(), 'error' );

				return array(
					'result'   => 'failure',
					'redirect' => ''
				);
			}

			// TODO: add optional call to Google Tag action
			return array(
				'result'   => 'success',
				'redirect' => 'javascript:window.flywire.WooCommerce.show(' . json_encode($config) . ')'
			);
		}
	} // end \WC_Gateway_Flywire class
}
