<?php
/**
 * Handle Flywire real-time payment callbacks
 */

if ( ! defined( 'ABSPATH' ) ) {
	exit;
}

class WC_Gateway_Flywire_Callback_Handler {
	/**
	 * @var bool
	 */
	private $complete_order;
	/**
	 * @var bool
	 */
	private $order_notes;
	/**
	 * @var bool
	 */
	private $on_guaranteed;

	/**
	 * Constructor.
	 *
	 * @param WC_Gateway_Flywire $gateway WC_Gateway_Flywire instance
	 */
	public function __construct( WC_Gateway_Flywire $gateway ) {
		add_action( 'woocommerce_api_wc_gateway_flywire', array( $this, 'check_response' ) );
		//add_action( 'valid_flywire_callback_received', array( $this, 'valid_response' ) );

		$this->complete_order = $gateway->complete_order;
		$this->order_notes    = $gateway->order_notes;
		$this->on_guaranteed  = $gateway->on_guaranteed;
	}

	/**
	 * Check for valid real-time callback response
	 */
	public function check_response() {
		$data = json_decode( file_get_contents( 'php://input' ), true );

		if ( ! empty( $data ) ) {
			$posted = wp_unslash( $data );
			error_log( 'Received Flywire callback: ' . json_encode( $posted ) );

			if ( $this->validate_callback( $posted ) ) {
				$this->valid_response( $posted );
				exit;
			}
		}

		wp_die( 'Flywire Callback Handler Failure', 'Flywire Callback', array( 'response' => 500 ) );
	}

	/**
	 * Check callback body has expected format
	 *
	 * @param array $response Post data after wp_unslash
	 *
	 * @returns bool
	 */
	public function validate_callback( array $response ): bool {
		$keys_to_check = array( "id", "amount", "status", "at", "callback_id" );
		$valid         = true;

		foreach ( $keys_to_check as $key ) {
			if ( ! isset( $response[ $key ] ) ) {
				$valid = false;
				break;
			}
		}

		return $valid;
	}

	/**
	 * There was a valid response, route to correct status method
	 *
	 * @param array $posted Post data after wp_unslash
	 */
	public function valid_response( array $posted ) {
		$order = wc_get_order( $posted['callback_id'] );

		if ( $order ) {
			$method_name = 'payment_status_' . $posted['status'];
			error_log( 'Found order #' . $order->get_id() . ' with current status: ' . $order->get_status() );

			if ( method_exists( $this, $method_name ) ) {
				error_log( 'Attempting to call method ' . $method_name . '()' );
				call_user_func( array( $this, $method_name ), $order, $posted );
			}
		}
	}

	/**
	 * Check if Order Notes setting is enabled and if so, add prettified order note
	 *
	 * @param WC_Order $order WC Order object
	 * @param array $posted Post data
	 * @param bool $create If true, add order note, otherwise just return formatted string
	 *
	 * @returns string
	 */
	protected function add_flywire_order_note( WC_Order $order, array $posted, $create = true ): string {
		$note = sprintf( __( 'Flywire transaction "%s" was %s on %s.', 'flywire-payment-gateway' ), $posted['id'], $posted['status'], $posted['at'] );

		if ( $this->order_notes && $create ) {
			$order->add_order_note( $note );
		}

		return $note;
	}

	/**
	 * Process payment status "initiated"
	 *
	 * @param WC_Order $order WC Order object
	 * @param array $posted callback request body
	 */
	protected function payment_status_initiated( WC_Order $order, array $posted ) {
		if ( $order->has_status( wc_get_is_paid_statuses() ) ) {
			WC_Gateway_Flywire::log( 'Aborting, Order #' . $order->get_id() . ' is already complete.' );
			exit;
		}

		if ( isset( WC()->cart ) ) {
			WC()->cart->empty_cart();
		}

		$this->payment_on_hold( $order );
		$this->add_flywire_order_note( $order, $posted );
	}

	/**
	 * Process payment status "guaranteed"
	 *
	 * @param WC_Order $order WC Order object
	 * @param array $posted callback request body
	 *
	 * @throws WC_Data_Exception
	 * @throws Exception
	 */
	protected function payment_status_guaranteed( WC_Order $order, array $posted ) {
		if ( $order->has_status( wc_get_is_paid_statuses() ) ) {
			WC_Gateway_Flywire::log( 'Aborting, Order #' . $order->get_id() . ' is already complete.' );
			exit;
		}

		$order_total = $order->get_total();
		$paid_amount = floatval( $posted['amount'] );

		if ($paid_amount < $order_total) {
			$order_item_id = wc_add_order_item( $order->get_id(), array(
				'order_item_name' => 'Partial payment',
				'order_item_type' => 'fee',
			));

			if ($order_item_id) {
				wc_add_order_item_meta( $order_item_id, '_fee_amount', $paid_amount*-1, true );
				wc_add_order_item_meta( $order_item_id, '_line_total', $paid_amount*-1, true );
				$order->calculate_totals();
				$order->save();
				$order->add_order_note( __( 'Partial payment of ' . get_woocommerce_currency_symbol( $order->get_currency() ) . $posted['amount'] . ' received', 'flywire-payment-gateway') );
			}
		} else if ( $this->on_guaranteed ) {
			$this->payment_complete( $order, $posted['id'], $this->add_flywire_order_note( $order, $posted, false ) );
		} else {
			$order->update_status( 'processing', $this->add_flywire_order_note( $order, $posted, false ) );
		}
	}

	/**
	 * Process payment status "delivered"
	 *
	 * @param WC_Order $order WC Order object
	 * @param array $posted callback request body
	 */
	protected function payment_status_delivered( WC_Order $order, array $posted ) {
		$this->add_flywire_order_note( $order, $posted );
		WC_Gateway_Flywire::log( 'Order #' . $order->get_id() . ' delivered.' );

		if ( ! $this->on_guaranteed ) {
			$this->payment_complete( $order, $posted['id'], $this->add_flywire_order_note( $order, $posted, false ) );
		}
	}

	/**
	 * Process payment status "cancelled"
	 *
	 * @param WC_Order $order WC Order object
	 * @param array $posted callback request body
	 */
	protected function payment_status_cancelled( WC_Order $order, array $posted ) {
		WC_Gateway_Flywire::log( 'Order #' . $order->get_id() . ' cancelled.' );
		$order->update_status( 'canceled', $this->add_flywire_order_note( $order, $posted, false ) );
	}

	/**
	 * Complete order, add transaction ID and note.
	 *
	 * @param WC_Order $order Order object
	 * @param string $txn_id Transaction ID
	 * @param string $note Payment note
	 */
	protected function payment_complete( WC_Order $order, $txn_id = '', $note = '' ) {
		if ( ! $order->has_status( array( 'processing', 'completed' ) ) ) {
			$order->add_order_note( $note );
			$order->payment_complete( $txn_id );

			WC_Gateway_Flywire::log( 'Order #' . $order->get_id() . ' completed with transaction ID "' . $txn_id . '"' );

			if ( $this->complete_order ) {
				$this->set_order_complete( $order );
			}
		}
	}

	/**
	 * Hold order and add note
	 *
	 * @param WC_Order $order WC Order object
	 * @param string $reason Reason why the payment is on hold
	 */
	protected function payment_on_hold( WC_Order $order, $reason = '' ) {
		$order->update_status( 'on-hold', $reason );

		if ( isset( WC()->cart ) ) {
			WC()->cart->empty_cart();
		}
	}

	/**
	 * Check if all products in order are virtual and,
	 * if true, set order status to complete
	 *
	 * @param WC_Order $order WC Order object
	 */
	protected function set_order_complete( WC_Order $order ) {
		$order_items = $order->get_items();

		// Create an array of products in the order
		$order_products = array_filter( array_map( function ( $item ) {
			// Get associated product for each line item
			return $item->get_product();
		}, $order_items ), function ( $product ) {
			// Remove non-products
			return ! ! $product;
		} );

		if ( count( $order_products ) > 0 ) {
			// Check if each product is 'virtual'
			$is_virtual_order = array_reduce( $order_products, function ( $virtual_order_so_far, $product ) {
				return $virtual_order_so_far && $product->is_virtual();
			}, true );

			if ( $is_virtual_order ) {
				$order->update_status( 'completed' );
				WC_Gateway_Flywire::log( 'Order #' . $order->get_id() . ' set to status Complete.' );
			} else {
				WC_Gateway_Flywire::log( 'Cannot set order #' . $order->get_id() . ' to Complete, as some products are not virtual.' );
			}
		}
	}
}