<?php
/**
 * Generate links to Flywire checkout page
 */

declare( strict_types=1 );

if ( ! defined( 'ABSPATH' ) ) {
	exit;
}

class WC_Gateway_Flywire_Request {
	/**
	 * WC order object
	 *
	 * @var WC_Order
	 */
	private $order;

	/**
	 * Field mapping definition object
	 *
	 * @var object
	 */
	private $mapping;

	/**
	 * Is in test mode.  If true, use UAT credentials and URLs.
	 *
	 * @var bool
	 */
	public $testmode;

	/**
	 * Flywire 3-letter portal code
	 *
	 * @var string
	 */
	public $portal;

	/**
	 * Webhook URL
	 *
	 * @var string
	 */
	public $api_url;

	/**
	 * Order return URL
	 *
	 * @var string
	 */
	public $return_url;


	/**
	 * Constructor.
	 *
	 * @param WC_Gateway_Flywire $gateway Flywire payment gateway
	 * @param string $order_id WC order ID
	 */
	public function __construct( WC_Gateway_Flywire $gateway, string $order_id ) {
		$this->order      = wc_get_order( $order_id );
		$this->mapping    = json_decode( $gateway->field_mapping );
		$this->portal     = apply_filters( 'flywpg_portal_code', $gateway->portal, $this->order, $gateway->testmode );
		$this->testmode   = $gateway->testmode;
		$this->api_url    = $gateway->api_url;
		$this->return_url = $gateway->get_return_url( $this->order );
	}

	/**
	 * Generate JSON config for Flywire Embed 2.0
	 *
	 * @return array
	 * @throws Exception
	 */
	public function get_config(): string {
		// Serialize order info into JSON
		$settings                = array();
		$settings['provider']    = 'woocommerce';
		$settings['env']         = $this->testmode ? 'demo' : 'production';
		$settings['locale']      = substr( $_SERVER['HTTP_ACCEPT_LANGUAGE'], 0, 2 );
		$settings['recipient']   = $this->portal;
		$settings['amount']      = intval( $this->order->get_total() * 100 );
		$settings['return_url']  = $this->return_url;
		$settings['callback_id'] = $this->order->get_id();
		$settings['callback_url'] = $this->api_url;
		foreach ( $this->mapping as $key => $value ) {
			$settings[ $key ] = $this->expand_template( $value, $this->order );
		}

		// Validate required fields are set
		$required_fields = array(
			'sender_country',
			'sender_first_name',
			'sender_last_name',
			'sender_address1',
			'sender_city',
			'sender_zip',
			'sender_email'
		);
		foreach ( $required_fields as $key ) {
			if ( ! isset( $settings[ $key ] ) ) {
				//wp_die( __( 'Embed 2.0 settings object is missing required parameter: ' . $key ) );
				throw new Exception( __( 'Embed 2.0 settings object is missing required parameter: ' . $key, 'flywire-payment-gateway' ) );
			}
		}

		return json_encode($settings);
	}


	/**
	 * Encode string to Base64URL
	 *
	 * @param string $data
	 *
	 * @return string
	 */
	public static function base64url_encode( string $data ): string {
		$b64 = base64_encode( $data );
		if ( $b64 === false ) {
			return '';
		}
		$url = strtr( $b64, '+/', '-_' );

		return rtrim( $url, '=' );
	}

	/**
	 * Generates a JSON key/value pair from the mapping template
	 *
	 * @param string $key property name
	 * @param string $template property value template
	 *
	 * @return string
	 */
	protected function map_field( string $key, string $template ): string {
		$json  = ',"' . $key . '":"';
		$order = $this->order;

		do {
			preg_match_all( '/{{([a-z_]+)}}/', $template, $matches, PREG_SET_ORDER );

			if ( count( $matches ) ) {
				$value    = call_user_func( array( $order, 'get_' . $matches[0][1] ) );
				$template = str_replace( $matches[0][0], $value, $template );
			}
		} while ( count( $matches ) );

		$json .= $template . '"';

		return $json;
	}

	/**
	 * Replaces tokens - e.g. {{name}} - in a template string with values from a given object
	 * Use {{custom.name}} for custom fields
	 *
	 * @param string $template template string
	 * @param object $obj Any object with getter methods to be used in template string
	 *
	 * @return string
	 */
	protected static function expand_template( string $template, $obj ): string {
		$re = '/{{(custom\.)?([a-z0-9_.]+)}}/';
		preg_match_all( $re, $template, $matches, PREG_SET_ORDER );

		while ( count( $matches ) ) {
			if ( 'custom.' == $matches[0][1] ) { // custom field in metadata
				$key = substr( $matches[0][2], 0, 1 ) === "_" ? $matches[0][2] : '_' . $matches[0][2];
				$value = $obj->get_meta( $key );
				if ('' == $value)
					$value = $obj->get_meta( ltrim($key, '_') ); // try without underscore
			} else { // native WC Order field
				if ( ! method_exists( $obj, 'get_' . $matches[0][2] ) ) {
					wc_add_notice( __( 'Field mapping failed: Method get_' . $matches[0][2] . '() not found.' ) );
				}
				$value = call_user_func( array( $obj, 'get_' . $matches[0][2] ) );
			}
			$template = str_replace( $matches[0][0], $value, $template );

			preg_match_all( $re, $template, $matches, PREG_SET_ORDER );
		}

		return $template;
	}
}
